# mkHttpd 1.0
# -----------
#
# Please see the web pages for releases and documentation.
#
# Author: Michael Kraus
#         mailto:mmg_kraus@csi.com
#         http://ourworld.compuserve.com/homepages/mmg_kraus
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted.
# The author makes no representations about the suitability of this
# software for any purpose.  It is provided "as is" without express
# or implied warranty.  By use of this software the user agrees to
# indemnify and hold harmless the author from any claims or
# liability for loss arising out of such use.

# we need some packages. with tcl 8.4 and its oo system, these will go.
package require mkGeneric
package require mkClasses

package provide mkHttpd 1.0

# name: Httpd
# args: args: option-value pairs
# the very simple http server class. accepts some options to configure the
# server socket and other things (see below). then creates the server socket.
# for each incoming connection, this class creates a Session (or whatever
# class is specified by the -session option) object that does all the actual
# work, like reading from the socket and responding. allowed options here are:
# -address ip-address: useful if the computer has more then one network card
# -port integer      : the tcp port the server will be listening to
# -rootdir path      : the root directory of the html etc. files
# -default filename  : the file to take, if the URL specifies a directory
# -session classname : the class to create an object of for each connect
class Httpd { args } {
  options aOpt $args {-address *} {-port * 80} {-name * mkHttpd-1.0} {-rootdir * .} {-default * index.html} {-session * Session} {-timeout * 120}

  my sName    $aOpt(-name)
  my sRootdir $aOpt(-rootdir)
  my sSession $aOpt(-session)
  my sDefault $aOpt(-default)
  my iTimeout [expr $aOpt(-timeout)*1000]

  if { ! [info exists aOpt(-address)] } {
    my hSock [socket -server "[this] _acceptConnect" $aOpt(-port)]
  } else {
    my hSock [socket -server "[this] _acceptConnect" -myaddr $aOpt(-address) $aOpt(-port)]
  }

  return [this]
} {
  if { [ismine hSock] } {
    close $hSock
  }
}

# name: Httpd:_acceptConnect
# args: hSock: socket that is connecting
#       sAddress: ip-address from where it is connecting
#       iPort: negotiated port used for this socket
# called whenever a connection is made. computes a unique name for a session
# object. creates that object if it doesn't exist. then resets the object
# so that it is ready to accept http requests. session objects are created
# per ip-address. that is, if a client has two browsers open and both connect
# to this server, then there will still be only one session object.
member Httpd:_acceptConnect { hSock sAddress iPort } {
  set sObject "[this].$sAddress"

  if { ! [cinfo isobject $sObject] } {
    [my sSession] $sObject $sAddress [my sRootdir] [my sDefault] [my sName] [my iTimeout]
  }

  $sObject _resetState $hSock
}

# name: Session
# args: sAddress: the client's ip-address
#       sRootdir: root directory of html etc. files
#       sDefault: file to take if URL specifies directory
#       sName   : server name to use in http responds
# the session class does all the work. it handles all the http requests
# coming from one particular client node and responds accordingly.
# each object creates an interpreter that remains state between the client's
# requests. this interpreter is used to evaluate the tcl code as compiled
# from the tsp files. inside the interpreter, the 'tsp' command provides a
# backdoor to its session object, so its methods can be called.
# usually, the session class is used as a superclass. the inheriting class
# would implement all functions required to retrieve application data that
# is to be displayed as a web page, hereby using tsp to define these pages.
# a tsp page can *only* mess around with its interpreter but not affect
# the application. also, it can *only* access the application by means
# of the members defined in this class or its inheriting class.
class Session { sAddress sRootdir sDefault sName iTimeout } {
  # on first usage, define common mime types and http error codes.
  # this can be easily extended if necessary.
  if { ! [isours bInit] } {
    our aMime.txt  text/plain
    our aMime.htm  text/html
    our aMime.html text/html
    our aMime.rtx  text/richtext
    our aMime.gif  image/gif
    our aMime.jpg  image/jpeg
    our aMime.jpeg image/jpeg
    our aMime.tiff image/tiff
    our aMime.tif  image/tiff
    our aMime.cgi  application/x-cgi
    our aMime.map  application/x-imagemap
    our aMime.xls  application/ms-excel
    our aMime.xlw  application/ms-excel
    our aMime.doc  application/msword
    our aMime.mdb  application/x-msaccess
    our aMime.wmf  application/x-msmetafile
    our aMime.exe  application/octet-stream
    our aMime.pdf  application/pdf
    our aMime.eps  application/postscript
    our aMime.ps   application/postscript
    our aMime.tcl  application/x-tcl
    our aMime.tex  application/x-tex
    our aMime.zip  application/zip
    our aMime.au   audio/basic
    our aMime.snd  audio/basic
    our aMime.wav  audio/x-wav
    our aMime.mp3  audio/mpeg
    our aMime.mpeg video/mpeg
    our aMime.mpg  video/mpeg
    our aMime.mov  video/quicktime
    our aMime.avi  video/x-msvideo
    our aMime.wmf  windows/metafile
    our aMime.wml  text/vnd.wap.wml
    our aMime.wmlc application/vnd.wap.wmlc
    our aMime.wbmp image/vnd.wap.wbmp
    our aMime.wmls text/vnd.wap.wmlscript
    our aMime.wmlsc application/vnd.wap.wmlscriptc

    our aCode.100 {100 Continue}
    our aCode.200 {200 Data follows}
    our aCode.204 {204 No Content}
    our aCode.302 {302 Found}
    our aCode.304 {304 Not Modified}
    our aCode.400 {400 Bad Request}
    our aCode.401 {401 Authorization Required}
    our aCode.403 {403 Permission denied}
    our aCode.404 {404 Not Found}
    our aCode.408 {408 Request Timeout}
    our aCode.411 {411 Length Required}
    our aCode.419 {419 Expectation Failed}
    our aCode.500 {500 Server Internal Error}
    our aCode.503 {503 Service Unavailable}
    our aCode.504 {504 Service Temporarily Unavailable}

    our bInit 1
  }

  # we need these parameters later, so they go into private variables
  my sName    $sName
  my sAddress $sAddress
  my sRootdir $sRootdir
  my sDefault $sDefault
  my iTimeout $iTimeout

  # this interp is used to evaluate all tsp code in. the alias opens a
  # backdoor from the interp to the object, but only to its methods. also,
  # some basic information about this session is stored in the Session array.
  # the exit command is redefined, so that it accepts an http return code
  # and raises an error. the error is caught by the tsp evaluating member.
  my hInterp  [interp create]
  [my hInterp] alias tsp [this]
  [my hInterp] eval [list array set Session [list ipaddress $sAddress rootdir $sRootdir default $sDefault]]
  [my hInterp] eval { proc exit { {iCode 200} args } { return -code error -errorcode HTTP "$iCode $args" } }
} {
  # destructor. all open sockets are closed.
  foreach sVar [mine *.hSock] {
    catch { close [my $sVar] }
  }

  # and the interp is deleted.
  [my hInterp] delete
}

# name: Session:_getDate
# args: iTime: a time in seconds
# helper function. returns a date/time string in http standard format. used
# in several mime keys. if iTime is empty, the current time is taken.
member Session:_getDate { {iTime {}} } {
  if { $iTime == {} } {
    set iTime [clock seconds]
  }

  return [clock format $iTime -format {%a, %d %b %Y %T %Z}]
}

# name: Session:_getMimeType
# args: sFile: any filename
# helper function. returns the mime type for the specified filename.
# looks up the aMime "array" to find the file's extension and returns
# the associated mime type. if not found, returns a default mime type.
member Session:_getMimeType { sFile } {
  set sExt [file extension $sFile]

  if { [isours aMime$sExt] } {
    return [our aMime$sExt]
  } else {
    return text/plain
  }
}

# name: Session:_getAlive
# args: hSock: a socket handle
# this helper function returns the value for the "connection" mime key.
# see Session:_resetState for the meaning of hSock.
member Session:_getAlive { hSock } {
  if { [ismine $hSock.mime.connection] } {
    return [my $hSock.mime.connection]
  } elseif { [my $hSock.sVersion] == "1.0" } {
    return "Close"
  } else {
    return "Keep-Alive"
  }
}

# name: _resetState
# args: hSock: a socket handle
# each socket handle represents an open connection to a client, so hSock can
# be seen as a token for a http connection. data specific to a connection is
# stored in private variables which are all named after the token, e.g.
# $hSock.sVersion.
# this function sets a connection back to its original state, so it gets ready
# to receive an http GET or POST request from the client. the first timer
# ensures that the socket is closed after 30s of inactivity. the second timer
# deletes the entire session object, if no valid http request has been received
# for more than 1 min.
member Session:_resetState { hSock } {
  # delete all data from previous requests over that socket
  eval unmy [mine $hSock.*]

  # store the token, set the state and save the timer handle
  my $hSock.hSock  $hSock
  my $hSock.sState wait
  my $hSock.hTimer [after 30000 "[this] _closeSocket $hSock"]

  # ready to receive ASCII data and process it in _readData
  fconfigure $hSock -blocking 0 -translation {auto crlf}
  fileevent $hSock readable "[this] _readData $hSock"

  # one timer per object, kills the object after iTimeout seconds
  if { ! [ismine hTimer] } {
    my hTimer [after [my iTimeout] "~[cinfo class [this]] [this]"]
  }
}

# name: Session:_closeSocket
# args: hSock: a socket handle
# closes a socket and deletes all related information of it.
member Session:_closeSocket { hSock } {
  catch { close $hSock }
  eval unmy [mine $hSock.*]
}

# name: Session:_eventuallyClose
# args: hSock: a socket handle
# either closes a socket or resets its state. this is called after a completed
# http transaction. the socket is closed either if the client closed its own
# end (flush would fail then), or if keep-alive isn't possible. otherwise the
# socket is reset to receive more http requests.
member Session:_eventuallyClose { hSock } {
  if { [catch {flush $hSock}] } {
    [this] _closeSocket $hSock
  } elseif { [[this] _getAlive $hSock] == "Close" } {
    [this] _closeSocket $hSock
  } else {
    [this] _resetState $hSock
  }
}

# name: Session:_readData
# args: hSock: a socket handle
# reads an http request. called whenever the socket gets readable. the value
# of $hSock.sState controls what part of the request is currently expected to
# be received: the GET or POST header, any number of mime keys, and eventually
# some post data. the latter is read as binary, the others line by line
# as ASCII. an http error is returned immediately whenever something
# unexpected is received. otherwise, an http response is initiated.
member Session:_readData { hSock } {
  # if we are currently reading post data (could be a lot and in binary)
  # then use 'read' to get it. otherwise 'gets'.
  if { [my $hSock.sState] == "post" } {
    set sLine [read $hSock]
    set iBytes [string length $sLine]
  } elseif { [catch {gets $hSock sLine} iBytes] } {
    [this] _closeSocket $hSock
    return
  }

  # a rip-off from the 'official' httpd package. the first part of the switch
  # pattern indicates if bytes were read (1), or not (0), or if we are blocked
  # or see an eof (-1). the second part is the state that indicates what part
  # of the http header is expected: the GET/POST line (wait), mime keys
  # (mime), or post data (post) if the http method is 'POST'.
  switch -glob -- [expr min($iBytes,1)].[my $hSock.sState] {
    1.wait {
      # first, we expect something like 'GET /login.html?user=me HTTP/1.1'
      if { ! [regexp {^(POST|GET) +([^?]+)\??([^ ]*) +HTTP/(1.[01])} $sLine sNix sMethod sUrl sQuery sVersion] } {
        [this] _replyError $hSock 400 "Garbled HTTP header"
        return
      }

      # store it for later use
      my $hSock.sMethod  [string toupper $sMethod] ;# method, GET or POST
      my $hSock.sUrl     $sUrl                     ;# the plain URL
      my $hSock.sQuery   $sQuery                   ;# query parameter string
      my $hSock.sVersion $sVersion                 ;# http version, 1.0 or 1.1

      # now we expect a bunch of mime keys, e.g. 'Connection: Keep-Alive'
      my $hSock.sState   mime
    }

    0.wait {
      # no bytes received. this can happen, doesn't matter.
      return
    }

    1.mime {
      # a mime key. we expect lines like 'key: value'
      if { ! [regexp {^([^ :]+):[ \t]*(.*)} $sLine sNix sKey sValue] } {
        [this] _replyError $hSock 400 "Garbled MIME data"
        return
      }

      # store it. some of them we need later
      my $hSock.mime.[string tolower $sKey] $sValue
    }

    0.mime {
      # this is the empty line after the mime keys. for method GET we're
      # done and can send the response back. for POST, we still have to
      # read the post data. the mime key 'Content-Length' tells us how many
      # bytes we still have to read.
      if { [my $hSock.sMethod] == "POST" } {
        if { ! [ismine $hSock.mime.content-length] } {
          [this] _replyError 411 "Mime key 'Content-Length' missing"
        }

        fconfigure $hSock -translation {binary crlf}
        my $hSock.iBytes [my $hSock.mime.content-length]
        my $hSock.sState post
      } else {
        [this] _processRequest $hSock
      }
    }

    1.post {
      # this is post data. in many cases this is in the same format as
      # the query parameters, so it is stored in the same variable. then
      # we keep track of how many bytes are still to be read.
      myappend $hSock.sQuery $sLine
      myincr $hSock.iBytes -$iBytes

      # done, when all bytes are received
      if { [my $hSock.iBytes] <= 0 } {
        [this] _processRequest $hSock
      }
    }

    0.post {
      # not sure if this should be here...
      [this] _processRequest $hSock
    }

    -1.* {
      # either got blocked or closed. for eof, the socket is closed.
      if { [eof $hSock] } {
        [this] _closeSocket $hSock
      }
    }

    default {
      # undefined state. should not happen.
      [this] _replyError $hSock 500 "Undefined server state"
      [this] _closeSocket $hSock
    }
  }
}

# name: Session:_processRequest
# args: hSock: a socket handle
# processes an http request. returns an http response to the client: either
# an error message, or the requested resource. if the resource specifies a
# tsp file then it is evaluated first, then the result is sent back.
member Session:_processRequest { hSock } {
  # cancel the two timers. the first would have closed the socket only,
  # the second would have killed the entire session object.
  catch { after cancel [my $hSock.hTimer] }
  catch { after cancel [my hTimer]; unmy hTimer }

  # get the 'real' filename
  set sFile [my sRootdir][my $hSock.sUrl]

  # if this is a directory, append the default page (e.g. index.html)
  if { [file isdirectory $sFile] } {
    append sFile /[my sDefault]
  }

  # this sequence checks if the file can be read and sends it back, *unless*
  # it is a *.tsp file. in this case the interp is updated with request-
  # specific data (e.g. the query params), then the file is parsed and
  # evaluated. the result is supposed to be a) an 'o.k.' (200) and a filename
  # containing the resulting html. the file is sent back to the client.
  # or b) a redirect (302) and a URL. the client gets redirected then.
  # or c) any http error code and some string describing the error.
  # or d) a zero (0) to do nothing. in this case the object itself MUST
  # create an http response, e.g. by using _replyData.
  if { ! [file exists $sFile] } {
    [this] _replyError $hSock 404 "Requested file not found"
    return
  } elseif { [file isdirectory $sFile] } {
    [this] _replyError $hSock 503 "Directory service not implemented"
    return
  } elseif { ! [file readable $sFile] } {
    [this] _replyError $hSock 403 "No read access for this file"
    return
  } elseif { [file extension $sFile] != ".tsp" } {
    [this] _replyFile $hSock $sFile
    return
  } elseif { [catch { [this] _setInterpEnv $hSock } sError] } {
    [this] _replyError $hSock 500 $sError $::errorInfo
    return
  } elseif { [catch { [this] _convertTspFile $sFile } lResult] } {
    [this] _replyError $hSock 500 $lResult $::errorInfo
    return
  } else {
    set iCode [lindex $lResult 0]
    set lInfo [lrange $lResult 1 end]

    switch $iCode {
      0       { if { [catch { eval [this] $lInfo } sError] } { [this] _replyError $hSock 500 $sError $::errorInfo } }
      200     { [this] _replyFile     $hSock $lInfo }
      302     { [this] _replyRedirect $hSock $lInfo }
      default { [this] _replyError    $hSock $iCode $lInfo }
    }
  }
}

# name: Session:_setInterpEnv
# args: hSock: a socket handle
# prepares the interp for evaluation of a tsp file. there are always four
# global arrays available: Session (set in the constructor), Request,
# Mime and Query.
member Session:_setInterpEnv { hSock } {
  # remember the socket. members called from within tsp files can use it,
  # but the tsp interp doesn't need to know it.
  my hSock $hSock

  # Request contains the http method (GET/POST), the URL and the query string
  set lR [list method [my $hSock.sMethod] url [my $hSock.sUrl] query [my $hSock.sQuery]]

  # Mime contains all mime keys and their values
  set lM {}
  foreach sKey [mine $hSock.mime.*] {
    lappend lM [lindex [split $sKey .] end] [my $sKey]
  }

  # Query contains the query parameters and their values
  # the query string (e.g. 'a=b&c=d+e') is split up and decoded.
  set lQ {}
  foreach sPair [split [my $hSock.sQuery] &] {
    foreach {sName sValue} [split $sPair =] {
      lappend lQ [[this] _decodeQuery $sName] [[this] _decodeQuery $sValue]
    }
  }

  # set these tree arrays
  [my hInterp] eval "
    catch { unset Request Mime Query }
    array set Request [list $lR]
    array set Mime    [list $lM]
    array set Query   [list $lQ]
  "
}

# name: Session:_decodeQuery
# args: sString: a URL encoded string
# a URL encoded string as in the query string of an http request does not
# contain blanks or other special characters, but rather + and escape
# sequences. this function restores the "original" string and returns it.
member Session:_decodeQuery { sString } {
  regsub -all {\+} $sString " " sString
  regsub -all {([][$\\])} $sString {\\\1} sString
  regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $sString  {[format %c 0x\1]} sString

  return [subst $sString]
}

# name: Session:_convertTspFile
# args: sTspFile: a tsp filename
# parses and evaluates a tsp file (tsp: tcl server pages). returns a list with
# two elements: a return code and a string. depending on the return code, the
# string is interpreted in different ways by the calling function:
# a) code 200: successful conversion of tsp to hmtl. the string is the created
# html file. b) code 302: the tsp evaluation returned a 'redirect'. the string
# is the new URL. c) anything else: as returned by the tsp script. the
# string is additional info defined by the tsp script.
# a tsp file normally contains html and tcl. all tcl must be enclosed within
# <% %>, hence it is similar to Microsoft's ASP concept. this function parses
# the tsp file and transforms it into a tcl script. the  script is evaluated,
# its result is html. both the tcl script and the html are written to disk, in
# a temporary directory '.tsp' in the same directory where the tsp file is.
# the tcl script file can so be reused next time, unless the tsp file has a
# newer time stamp. the html file is later sent to the client with fcopy
# (in the background). since the resulting html can vary from client to client,
# the file name for the html file contains the client's ip address.
# note: obviously, a tsp file can contain any markup, like xml, wml, html...
member Session:_convertTspFile { sTspFile } {
  # determine the temp directory and the filenames for the tcl and html files
  set sDir [file dirname $sTspFile]/.tsp
  set sRootname $sDir/[file rootname [file tail $sTspFile]]
  set sTclFile  $sRootname.tcl
  set sHtmlFile $sRootname.[my sAddress].html

  # if the tcl file already exists and is newer than the tsp file, then
  # parsing of the tsp file is not necessary.
  if { [file exists $sTclFile] && [file mtime $sTclFile] > [file mtime $sTspFile] } {
    # the tcl script is up-to-date. take it directly from the file.
    set hTclFile [open $sTclFile r]
    set sCmd [read $hTclFile]
    close $hTclFile
  } else {
    # parsing required. first read the tsp file.
    set hTspFile [open $sTspFile r]
    set sTsp "  [read $hTspFile]"
    close $hTspFile

    # main part of the script is a string '__sHtml__'. all html code from the
    # tsp (i.e. everything not within <% %> brackets) is appended to it,
    # considering any tcl (inside <% %>) as surrounding control code.
    set sCmd "set __sHtml__ {
    <!-- ------------------------------------------------------------- --
      -- Autogenerated on [clock format [clock seconds] -format {%D %T}] by mkHttpd 1.0 and TSP 1.0 --
      -- ------------------------------------------------------------- -->
    }\n"

    # now find all <% %> brackets. inside we expect tcl, outside html.
    set iStop 0
    while 1 {
      # get a chunk of html
      set iStart [string first "<%" $sTsp $iStop]
      if { $iStart == -1 } break
      set sHtml [string range $sTsp [expr $iStop + 2] [expr $iStart - 1]]

      # get a chunk of tcl
      set iStop [string first "%>" $sTsp $iStart]
      if { $iStop == -1 } break
      set sTcl [string range $sTsp [expr $iStart + 2] [expr $iStop - 1]]

      # append the html to the sHtml string of the script, then append the
      # tcl to the script directly.
      append sCmd "append __sHtml__ \[subst {$sHtml}\]\n" "$sTcl\n"
    }

    # no more <% %> found. append some remaining html to the script.
    set sHtml [string range $sTsp [expr $iStop + 2] end]
    append sCmd "append __sHtml__ \[subst {$sHtml}\]\n"

    # now force the creation of the .tsp temp directory
    if { ! [file isdirectory $sDir] } {
      file delete $sDir
      file mkdir $sDir
    }

    # and write the tcl script file.
    set hTclFile [open $sTclFile w]
    puts $hTclFile $sCmd
    close $hTclFile
  }

  # now eval the script. if it fails and the error code is not HTTP, then
  # it is a true error. escalate it.
  if { ! [catch { [my hInterp] eval $sCmd } lResult] } {
    set lResult 200
  } elseif { $::errorCode != "HTTP" } {
    error $lResult $::errorInfo $::errorCode
  }

  # if the script has returned result code 200 (ok) then the evaluation
  # result is stored in an html file.
  if { $lResult == 200 } {
    set sHtml [[my hInterp] eval set __sHtml__]

    set hHtmlFile [open $sHtmlFile w]
    puts $hHtmlFile $sHtml
    close $hHtmlFile

    lappend lResult $sHtmlFile
  }

  return $lResult
}

# name: Session:_replyFile
# args: hSock: a socket handle
#       sFile: file to return contents of
# sends the requested resource back to the client. first sends a standard
# http header, then the file's contents via fcopy, in binary mode. fcopy
# does that in the background, so we will end up in _replyDone when the
# copy is done (or when the client closed the connection, e.g. because
# the browser decides to take the file from its cache).
member Session:_replyFile { hSock sFile } {
  puts $hSock "HTTP/[my $hSock.sVersion] [our aCode.200]"
  puts $hSock "Server: [my sName]"
  puts $hSock "Date: [[this] _getDate]"
  puts $hSock "Connection: [[this] _getAlive $hSock]"
  puts $hSock "Last-Modified: [[this] _getDate [file mtime $sFile]]"
  puts $hSock "Content-Type: [[this] _getMimeType $sFile]"
  puts $hSock "Content-Length: [file size $sFile]"
  puts $hSock ""

  set hFile [open $sFile]
  fconfigure $hFile -translation binary
  fconfigure $hSock -translation binary

  fcopy $hFile $hSock -command "[this] _replyDone $hSock $hFile"
}

# name: Session:_replyDone
# args: hSock: a socket handle
#       hFile: the file handle from _replyFile
#       iBytes: number of bytes transmitted
#       sError: any error that happened during transmission
# fcopy calls this after it completed the transfer. the file is closed,
# and the socket is flushed and eventually closed.
member Session:_replyDone { hSock hFile iBytes {sError "(none)"} } {
  close $hFile
  [this] _eventuallyClose $hSock
}

# name: Session:_replyError
# args: hSock: a socket handle
#       iCode: http code to be returned
#       sInfo: any verbose text to describe the reason
# returns an http response with an error code. the information in sInfo
# is formatted into a short html page and sent as well.
member Session:_replyError { hSock iCode {sReason "n/a"} {sInfo {}} } {
  if { ! [isours aCode.$iCode] } {
    set sInfo "Undefined reply code $iCode ('$sReason') created"
    set iCode 500
  }

  regsub -all {<} $sInfo {&lt;} sInfo
  regsub -all {<} $sInfo {&gt;} sInfo
  regsub -all {\n} $sInfo {<br>} sInfo

  set sText "
    <HTML>
      <HEAD>
        <TITLE>mkHttpd Error $iCode</TITLE>
      </HEAD>
      <BODY>
        <H2>mkHttpd Error '[our aCode.$iCode]'</H2><BR>
        Resource: [my $hSock.sUrl]<BR>
        Query: [my $hSock.sQuery]<BR>
        Reason: $sReason<BR><br>
        <pre><font size=\"1\" color=\"#888888\">$sInfo</font></pre>
        <BR>
      </BODY>
    </HTML>
  "

  puts $hSock "HTTP/[my $hSock.sVersion] [our aCode.$iCode]"
  puts $hSock "Server: [my sName]"
  puts $hSock "Date: [[this] _getDate]"
  puts $hSock "Connection: [[this] _getAlive $hSock]"
  puts $hSock "Content-Type: text/html"
  puts $hSock "Content-Length: [string length $sText]"
  puts $hSock ""
  puts -nonewline $hSock $sText

  [this] _eventuallyClose $hSock
}

# name: Session:_replyRedirect
# args: hSock: a socket handle
#       sUrl: URL to redirect the client to
# returns an http response with a redirection directive (http return code
# 302). The URL is embedded as a link into a small html page.
member Session:_replyRedirect { hSock sUrl } {
  set sText "
    <HTML>
      <HEAD>
        <TITLE>mkHttpd Redirection</TITLE>
      </HEAD>
      <BODY>
        <H2>Document has moved</H2><BR>
        This document has moved to <A HREF=\"$sUrl\">$sUrl</A>.
      </BODY>
    </HTML>
  "

  puts $hSock "HTTP/[my $hSock.sVersion] [our aCode.302]"
  puts $hSock "Server: [my sName]"
  puts $hSock "Date: [[this] _getDate]"
  puts $hSock "Connection: [[this] _getAlive $hSock]"
  puts $hSock "Content-Type: text/html"
  puts $hSock "Content-Length: [string length $sText]"
  puts $hSock "Location: $sUrl"
  puts $hSock ""
  puts -nonewline $hSock $sText

  [this] _eventuallyClose $hSock
}

# name: Session:_replyData
# args: hSock: a socket handle
#       sMime: a mime type, e.g. text/html
#       sData: data to be sent
member Session:_replyData { hSock sMime sData } {
  puts $hSock "HTTP/[my $hSock.sVersion] [our aCode.200]"
  puts $hSock "Server: [my sName]"
  puts $hSock "Date: [[this] _getDate]"
  puts $hSock "Connection: [[this] _getAlive $hSock]"
  puts $hSock "Content-Type: $sMime"
  puts $hSock "Content-Length: [string length $sData]"
  puts $hSock ""

  puts -nonewline $hSock $sData

  [this] _eventuallyClose $hSock
}

# name: Session:replyFile
# args: see _replyFile
# Callable from inside TSP. Considered protected.
member Session:replyFile { sFile } {
  [this] _replyFile [my hSock] $sFile
}

# name: Session:replyError
# args: see _replyError
# Callable from inside TSP. Considered protected.
member Session:replyError { iCode {sReason "n/a"} {sInfo {}} } {
  [this] _replyError [my hSock] $iCode $sReason $sInfo
}

# name: Session:replyRedirect
# args: see _replyRedirect
# Callable from inside TSP. Considered protected.
member Session:replyRedirect { sUrl } {
  [this] _replyRedirect [my hSock] $sUrl
}

# name: Session:replyData
# args: see _replyData
# Callable from inside TSP. Considered protected.
member Session:replyData { sMime sData } {
  [this] _replyData [my hSock] $sMime $sData
}
